###
### "THE BEER-WARE LICENSE":
### Alberto Ramos wrote this file. As long as you retain this 
### notice you can do whatever you want with this stuff. If we meet some 
### day, and you think this stuff is worth it, you can buy me a beer in 
### return. <alberto.ramos@cern.ch>
###
### file:    ADerrorsUtils.jl
### created: Fri Jun 26 19:30:27 2020
###                               

function root_error(fnew::Function, x::Float64,
                    data::Vector{uwreal}) 

    function fvec(x::Vector)
        return fnew(x[1], x[2:end])
    end
    xv = zeros(Float64, length(data)+1)
    xv[1] = x
    for i in 1:length(data)
        xv[i+1] = data[i].mean
    end
    grad = ForwardDiff.gradient(fvec, xv)

    return addobs(data, -grad[2:end]./grad[1], x)
end

function chiexp(hess::Array{Float64, 2}, data::Vector{uwreal}, W::Vector{Float64})

    m = length(data)
    n = size(hess, 1) - m

    hm = view(hess, 1:n, n+1:n+m)
    sm = hm * LinearAlgebra.Diagonal(1.0 ./ sqrt.(W))

    Px = LinearAlgebra.Symmetric(LinearAlgebra.Diagonal(W)) -
        LinearAlgebra.Symmetric(hm' *
                                LinearAlgebra.inv(LinearAlgebra.Symmetric(sm * sm')) *
                                hm)

    return trcov(Px, data)
end
    
function chiexp(hess::Array{Float64, 2}, data::Vector{uwreal}, W::Array{Float64, 2})

    m = length(data)
    n = size(hess, 1) - m
    
    Lm = LinearAlgebra.cholesky(LinearAlgebra.Symmetric(W))
    Li = LinearAlgebra.inv(Lm.L)
    
    hm = view(hess, 1:n, n+1:n+m)
    sm = hm * Li'
    
    Px = LinearAlgebra.Symmetric(W) -
        LinearAlgebra.Symmetric(hm' *
                                LinearAlgebra.inv(LinearAlgebra.Symmetric(sm * sm')) *
                                hm)
    
    return trcov(Px, data)
end

function chiexp(chisq::Function,
                xp::Vector{Float64}, 
                data::Vector{uwreal};
                W = Vector{Float64}())

    n = length(xp)   # Number of fit parameters
    m = length(data) # Number of data

    xav = zeros(Float64, n+m)
    for i in 1:n
        xav[i] = xp[i]
    end
    for i in n+1:n+m
        xav[i] = data[i-n].mean
    end
    ccsq(x::Vector) = chisq(view(x, 1:n), view(x, n+1:n+m))
    hess = Array{Float64}(undef, n+m, n+m)
#    @time ForwardDiff.hessian!(hess, ccsq, xav)
    hyperd_hessian!(hess, ccsq, xav)

    cse = 0.0
    if (m-n > 0)
        if (length(W) == 0)
            Ww = zeros(Float64, m)
            for i in 1:m
                if (data[i].err == 0.0)
                    uwerr(data[i])
                    if (data[i].err == 0.0)
                        error("Zero error in fit data")
                    end
                end
                Ww[i] = 1.0 / data[i].err^2
            end
        else
            Ww = W
        end

        cse = chiexp(hess, data, Ww)        
    end

    return cse
end

function fit_error(chisq::Function,
                   xp::Vector{Float64}, 
                   data::Vector{uwreal};
                   W = Vector{Float64}(), chi_exp = true)

    n = length(xp)   # Number of fit parameters
    m = length(data) # Number of data

    xav = Vector{Float64}(undef, n+m)
    for i in 1:n
        xav[i] = xp[i]
    end
    for i in n+1:n+m
        xav[i] = data[i-n].mean
    end

    ccsq(x::Vector) = chisq(view(x, 1:n), view(x, n+1:n+m))
    hess = Array{Float64}(undef, n+m, n+m)
#    @time ForwardDiff.hessian!(hess, ccsq, xav)
    hyperd_hessian!(hess, ccsq, xav)
    
    hinv = LinearAlgebra.pinv(hess[1:n,1:n])
    grad = - hinv[1:n,1:n] * hess[1:n,n+1:n+m]
    
    param = addobs(data, grad, xp)

    if (!chi_exp)
        return param
    end
    
    cse = 0.0
    if (m-n > 0)
        if (length(W) == 0)
            Ww = zeros(Float64, m)
            for i in 1:m
                if (data[i].err == 0.0)
                    uwerr(data[i])
                    if (data[i].err == 0.0)
                        error("Zero error in fit data")
                    end
                end
                Ww[i] = 1.0 / data[i].err^2
            end
        else
            Ww = W
        end
        
        cse = chiexp(hess, data, Ww)        
    end

    return param, cse
end

function int_error(fint::Function,
                   a,
                   b,
                   param::Vector{uwreal})

    # First basic integral
    xp = zeros(Float64, length(param))
    for i in 1:length(xp)
        xp[i] = param[i].mean
    end
    f(x) = fint(x, xp)
    (val, foo) = QuadGK.quadgk(f, a, b)

    # Functions that return the derivatives
    # with respect to the parameters as functions
    function fp(n::Int64)
        function fd(x)
            fax(pa) = fint(x,[xp[1:n-1];[pa];xp[n+1:end]])
            return ForwardDiff.derivative(fax, xp[n])
        end
        return fd
    end

    grad = zeros(Float64, length(param))
    for i in 1:length(param)
        (grad[i], foo) = QuadGK.quadgk(fp(i), a, b)
    end

    return addobs(param, grad, val)
end

function int_error(fint::Function,
                   a::uwreal,
                   b::Float64,
                   param::Vector{uwreal})

    # First basic integral
    xp = zeros(Float64, length(param))
    for i in 1:length(xp)
        xp[i] = param[i].mean
    end
    f(x) = fint(x, xp)
    (val, foo) = QuadGK.quadgk(f, a.mean, b)

    # Functions that return the derivatives
    # with respect to the parameters as functions
    function fp(n::Int64)
        function fd(x)
            fax(pa) = fint(x,[xp[1:n-1];[pa];xp[n+1:end]])
            return ForwardDiff.derivative(fax, xp[n])
        end
        return fd
    end

    grad = zeros(Float64, length(param)+1)
    for i in 1:length(param)
        (grad[i], foo) = QuadGK.quadgk(fp(i), a.mean, b)
    end
    grad[end] = fint(a.mean, xp)

    return addobs([param; a], grad, val)
end

function int_error(fint::Function,
                   a::uwreal,
                   b::uwreal,
                   param::Vector{uwreal})

    # First basic integral
    xp = zeros(Float64, length(param))
    for i in 1:length(xp)
        xp[i] = param[i].mean
    end
    f(x) = fint(x, xp)
    (val, foo) = QuadGK.quadgk(f, a.mean, b.mean)

    # Functions that return the derivatives
    # with respect to the parameters as functions
    function fp(n::Int64)
        function fd(x)
            fax(pa) = fint(x,[xp[1:n-1];[pa];xp[n+1:end]])
            return ForwardDiff.derivative(fax, xp[n])
        end
        return fd
    end

    grad = zeros(Float64, length(param)+2)
    for i in 1:length(param)
        (grad[i], foo) = QuadGK.quadgk(fp(i), a.mean, b.mean)
    end
    grad[end-1] = fint(a.mean, xp)
    grad[end]   = fint(b.mean, xp)

    return addobs([param; a; b], grad, val)
end

function int_error(fint::Function,
                   a::Float64,
                   b::uwreal,
                   param::Vector{uwreal})

    # First basic integral
    xp = zeros(Float64, length(param))
    for i in 1:length(xp)
        xp[i] = param[i].mean
    end
    f(x) = fint(x, xp)
    (val, foo) = QuadGK.quadgk(f, a, b.mean)

    # Functions that return the derivatives
    # with respect to the parameters as functions
    function fp(n::Int64)
        function fd(x)
            fax(pa) = fint(x,[xp[1:n-1];[pa];xp[n+1:end]])
            return ForwardDiff.derivative(fax, xp[n])
        end
        return fd
    end

    grad = zeros(Float64, length(param)+1)
    for i in 1:length(param)
        (grad[i], foo) = QuadGK.quadgk(fp(i), a, b.mean)
    end
    grad[end]   = fint(b.mean, xp)

    return addobs([param; b], grad, val)
end
